﻿''' <summary>
''' Author: Oliver Schirm
''' Date: 16.10.2015
''' Description: GUI for BasicOCR Class
''' </summary>
Public Class frmMain
#Region "Settings"
    Private sNeuralNetFileName As String = "Brain.dat"
    Private sTrainingPatternsFileName As String = "TrainingPatterns.dat"
    Private iTilesX As Integer = 5
    Private iTilesY As Integer = 10
    Private iHiddenNeuronsCount As Integer = 20
    Private dLearningRate As Double = 0.4
    Private dMomentum As Double = 0.2
    Private dMaxError As Double = 0.1
    Private bDebug As Boolean = False
#End Region
#Region "Global declarations"
    Private bPaintingEnabled As Boolean = True
    Private bPainting As Boolean = False
#End Region
#Region "GUI Modes"
    ''' <summary>
    ''' GUI mode for training of a wrong recognized letter.
    ''' </summary>
    Private Sub enterCorrectionMode()
        bPaintingEnabled = False
        btnWrong.Enabled = False
        txtLetter.Enabled = True
        btnOK.Enabled = True
        btnClear.Enabled = False
        lblWhichLetter.Enabled = True
        picPaint.Enabled = False
        prgTrainingProgress.Value = 0
    End Sub
    ''' <summary>
    ''' GUI mode for painting and letter recognition.
    ''' </summary>
    Private Sub enterPaintingMode()
        txtLetter.Enabled = False
        btnOK.Enabled = False
        bPaintingEnabled = True
        btnClear.Enabled = True
        lblWhichLetter.Enabled = False
        btnWrong.Enabled = True
        prgTrainingProgress.Value = 0
        picPaint.Enabled = True
    End Sub
    ''' <summary>
    ''' GUI mode for trainig. Everything disabled.
    ''' </summary>
    Private Sub enterTrainingMode()
        txtLetter.Enabled = False
        btnOK.Enabled = False
        bPaintingEnabled = False
        btnClear.Enabled = False
        lblWhichLetter.Enabled = False
        picPaint.Enabled = False
    End Sub
#End Region
#Region "GUI interaction handling"
    ''' <summary>
    ''' Enters correction mode.
    ''' </summary>
    Private Sub btnWrong_Click(sender As Object, e As EventArgs) Handles btnWrong.Click
        enterCorrectionMode()
    End Sub
    ''' <summary>
    ''' Enables the OK button if there is exactly 1 letter in textbox.
    ''' </summary>
    Private Sub txtLetter_TextChanged(sender As Object, e As EventArgs) Handles txtLetter.TextChanged
        If txtLetter.Text.Length = 1 Then
            btnOK.Enabled = True
        Else
            btnOK.Enabled = False
        End If
    End Sub
    ''' <summary>
    ''' Corrects wrong recognized letter.
    ''' </summary>
    Private Sub btnOK_Click(sender As Object, e As EventArgs) Handles btnOK.Click
        txtLetter.Text = txtLetter.Text.ToUpper
        enterTrainingMode()
        correctLetter()
        resetPaintingGround()
        enterPaintingMode()
    End Sub
    ''' <summary>
    ''' Opens VB-Paradise thread for Brain.dll.
    ''' </summary>
    Private Sub lnkVBParadise_LinkClicked(sender As Object, e As LinkLabelLinkClickedEventArgs) Handles lnkVBParadise.LinkClicked
        Process.Start("http://www.vb-paradise.de/index.php/Thread/114653-Brain-dll-Neuronale-Netze-f%C3%BCr-KI-s-und-Vorhersagen/")
    End Sub
    ''' <summary>
    ''' Clears the painting ground in order to paint a new letter.
    ''' </summary>
    Private Sub btnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        resetPaintingGround()
    End Sub
    ''' <summary>
    ''' Sets the progress bar to training progress value while training.
    ''' </summary>
    Private Shared Sub progressChanged(ByVal iPercent As Integer)
        frmMain.prgTrainingProgress.Value = iPercent
        Application.DoEvents()
    End Sub
#End Region
#Region "Painting on GUI"
    ''' <summary>
    ''' Starts painting on mouse down if painting is enabled.
    ''' </summary>
    Private Sub picPaint_MouseDown(sender As Object, e As MouseEventArgs) Handles picPaint.MouseDown
        If bPaintingEnabled = True Then
            bPainting = True
        End If
    End Sub
    ''' <summary>
    ''' Paints on mouse move if painting is enabled and mouse is down.
    ''' </summary>
    Private Sub picPaint_MouseMove(sender As Object, e As MouseEventArgs) Handles picPaint.MouseMove
        If bPainting = True Then
            Dim g As Graphics = Graphics.FromImage(picPaint.Image)
            g.FillEllipse(Brushes.Black, New System.Drawing.Rectangle(New System.Drawing.Point(e.X, e.Y), New System.Drawing.Size(New System.Drawing.Point(10, 10))))
            g.Dispose()
            picPaint.Invalidate()
        End If
    End Sub
    ''' <summary>
    ''' Stops painting on mouse up and does letter recognition.
    ''' </summary>
    Private Sub picPaint_MouseUp(sender As Object, e As MouseEventArgs) Handles picPaint.MouseUp
        bPainting = False
        recognizeLetter()
    End Sub
#End Region
#Region "Algorithms"
    ''' <summary>
    ''' Corrects wrong recognized letter by adding the painted letter to training sets and running the training again.
    ''' </summary>
    Private Sub correctLetter()
        lblTrainingPatterns.Text = "Training Patterns: " & BasicOCR.addTrainingSet(CType(picPaint.Image, Bitmap), txtLetter.Text.ToLower.ToCharArray()(0), iTilesX, iTilesY).ToString
        BasicOCR.trainTrainingSet(dMaxError)
    End Sub
    ''' <summary>
    ''' Resets the picture box.
    ''' </summary>
    Private Sub resetPaintingGround()
        Dim bmp As Bitmap = New Bitmap(picPaint.Width, picPaint.Height)
        picPaint.Image = bmp
        Dim g As Graphics = Graphics.FromImage(picPaint.Image)
        g.Clear(Color.White)
        g.Dispose()
    End Sub
    ''' <summary>
    ''' Recognizes the currently painted letter and shows the result in result Label.
    ''' </summary>
    Private Sub recognizeLetter()
        lblResult.Text = BasicOCR.getLetter(CType(picPaint.Image, Bitmap), iTilesX, iTilesY).ToString.ToUpper
    End Sub
#End Region
#Region "Program startup"
    ''' <summary>
    ''' Handles on program startup.
    ''' Initializes the painting ground and loads the training sets and neural net from file.
    ''' </summary>
    Private Sub frmMain_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        resetPaintingGround()
        enterPaintingMode()
        BasicOCR.init(iTilesX * iTilesY, iHiddenNeuronsCount, dLearningRate, dMomentum, bDebug)
        If My.Computer.FileSystem.FileExists(sTrainingPatternsFileName) Then
            Dim iTrainingSetsLoaded As Integer = BasicOCR.loadTrainingSets(sTrainingPatternsFileName)
            lblTrainingPatterns.Text = "Training Patterns: " & iTrainingSetsLoaded
            If My.Computer.FileSystem.FileExists(sNeuralNetFileName) Then
                BasicOCR.restoreBrain(sNeuralNetFileName)
            End If
        End If
        AddHandler BasicOCR.TrainingProgressChanged, AddressOf progressChanged
    End Sub
#End Region
#Region "Program shutdown"
    ''' <summary>
    ''' Handles on closing the main window.
    ''' Saves the current training sets and the neural net into a file.
    ''' </summary>
    Private Sub frmMain_FormClosing(sender As Object, e As FormClosingEventArgs) Handles Me.FormClosing
        BasicOCR.saveTrainingSets(sTrainingPatternsFileName)
        BasicOCR.saveBrain(sNeuralNetFileName)
    End Sub
#End Region
End Class
